local ipairs = ipairs
local print = print
local select = select
local strsplit = strsplit
local strtrim = strtrim
local string = string
local table = table
local _GetTime = GetTime

local loaded = false
local quantity = 0
local lastKill = {0} -- To be populated later, do not remove the initial value. The zero means inconclusive/invalid data.
local currentPullUpdateTimer = 0
local activeNameplates = {}

mppDebug = false

mppSimulationMode = false
local simulationMax = 220
local simulationCurrent = 28

local version = "0.3e"
local numVersion = 6
local defaultSettings = {
	version = version,
	dbVersion = numVersion,
	enabled = true,
	
	inconclusiveDataThreshold = 100, -- Mobs killed within this span of time (in milliseconds) will not be processed since we might not get the criteria update fast enough to know which mob gave what progress. Well, that's the theory anyway.
	maxTimeSinceKill = 600, -- Lag tolerance between a mob dying and the progress criteria updating, in milliseconds.
	
	enableTooltip = true,
	tooltipColor = "82E0FF", -- Color of the tooltip. Duh.
	
	enablePullEstimate = true,
	pullEstimateCombatOnly = false,
	
	nameplateUpdateRate = 200, -- Rate (in milliseconds) at which we update the progress we get from the current pull, as estimated by active name plates you're in combat with. Also the update rate of getting new values for nameplate text overlay if enabled.
	
	enableNameplateText = true,
	nameplateTextColor = "FFFFFF",
}

local warnings = {}

mppDefaultValues = {

	-- Test Mobs

	{42286, 50, "Young Snow Leopard"},

	-- Atal Dazar
	{127799, 4, "Dazar'ai Honor Guard"},
	{122971, 4, "Dazar'ai Juggernaut"},
	{122972, 4, "Dazar'ai Augur"},
	{122970, 4, "Shadowblade Stalker"},
	{129553, 10, "Dinomancer Kish'o"},
	{127757, 4, "Reanimated Honor Guard"},
	{128455, 6, "T'lonja"},
	{122984, 6, "Dazar'ai Colossus"},
	{122969, 4, "Zanchuli Witch-Doctor"},
	{122973, 4, "Dazar'ai Confessor"},
	{132126, 4, "Gilded Priestess"},
	{128435, 1, "Toxic Saurid"},
	{128434, 4, "Feasting Skyscreamer"},
	{129552, 6, "Monzumi"},

	-- Freehold

	{126919, 3, "Irontide Stormcaller"},
	{128551, 4, "Irontide Mastiff"},
	{126918, 4, "Irontide Crackshot"},
	{126928, 4, "Irontide Corsair"},
	{129788, 4, "Irontide Bonesaw"},
	{129598, 2, "Freehold Pack Mule"},
	{129559, 4, "Cutwater Duelist"},
	{129548, 4, "Blacktooth Brute"},
	{130024, 1, "Soggy Shiprat"},
	{129529, 4, "Blacktooth Scrapper"},
	{129527, 4, "Bilge Rat Buccaneer"},
	{130400, 6, "Irontide Crusher"},
	{130012, 4, "Irontide Ravager"},
	{127119, 1, "Freehold Deckhand"},
	{130522, 2, "Freehold Shipmate"},
	{129600, 3, "Bilge Rat Brinescale"},
	{127106, 6, "Irontide Officer"},
	{127124, 1, "Freehold Barhand"},
	{127111, 6, "Irontide Oarsman"},
	{130404, 4, "Vermin Trapper"},
	{129550, 4, "Bilge Rat Padfoot"},
	{130011, 4, "Irontide Buccaneer"},
	{129547, 4, "Blacktooth Knuckleduster"},
	{129526, 4, "Bilge Rat Swabby"},
	{129602, 6, "Irontide Enforcer"},
	{129599, 3, "Cutwater Knife Juggler"},
	{129601, 4, "Cutwater Harpooner"},

	-- Kings Rest
	{133935, 5, "Animated Guardian"},
	{134158, 6, "Shadow-Borne Champion"},
	{137486, 4, "Queen Patlaa"},
	{134739, 8, "Purification Construct"},
	{137969, 6, "Interment Construct"},
	{137484, 6, "King A'akul"},
	{134157, 4, "Shadow-Borne Warrior"},
	{137485, 4, "Bloodsworn Agent"},
	{135167, 4, "Spectral Berserker"},
	{135239, 4, "Spectral Witch Doctor"},
	{134251, 4, "Seneschal M'bara"},
	{135231, 8, "Spectral Brute"},
	{134174, 5, "Shadow-Borne Witch Doctor"},
	{137473, 4, "Guard Captain Atu"},
	{135192, 4, "Honored Raptor"},
	{135204, 4, "Spectral Hex Priest"},
	{134994, 1, "Spectral Headhunter"},
	{134331, 6, "King Rahu'ai"},
	{137487, 4, "Skeletal Hunting Raptor"},
	{137474, 6, "King Timalji"},
	{135235, 4, "Spectral Beastmaster"},
	{138489, 8, "Shadow of Zul"},
	{137478, 6, "Queen Wasi"},
	
	--Shrine of the Storm

	{134139, 10, "Shrine Templar"},
	{136347, 1, "Tidesage Initiate"},
	{134173, 1, "Animated Droplet"},
	{139800, 9, "Galecaller Apprentice"},
	{139626, 1, "Dredged Sailor"},
	{134338, 9, "Tidesage Enforcer"},
	{136214, 18, "Windspeaker Heldis"},
	{134418, 9, "Drowned Depthbringer"},
	{136186, 9, "Tidesage Spiritualist"},
	{139799, 9, "Ironhull Apprentice"},
	{136353, 10, "Colossal Tentacle"},
	{134423, 1, "Abyss Dweller"},
	{134150, 18, "Runecarver Sorn"},
	{134137, 9, "Temple Attendant"},
	{134144, 13, "Living Current"},
	{136249, 18, "Guardian Elemental"},
	{136295, 13, "Sunken Denizen"},
	{136297, 9, "Forgotten Denizen"},
	{134514, 9, "Abyssal Cultist"},
	{134417, 10, "Deepsea Ritualist"},
	{140038, 3, "Abyssal Eel"},

	--Siege of Boralus
	{137511, 4, "Bilge Rat Cutthroat"},
	{141495, 1, "Kul Tiran Footman"},
	{141284, 4, "Kul Tiran Wavetender"},
	{141285, 4, "Kul Tiran Marksman"},
	{141566, 1, "Scrimshaw Gutter"},
	{129640, 4, "Snarling Dockhound"},
	{129372, 4, "Blacktar Bomber"},
	{135258, 1, "Irontide Marauder"},
	{128969, 8, "Ashvane Commander"},
	{138464, 4, "Ashvane Deckhand"},
	{135245, 8, "Bilge Rat Demolisher"},
	{128967, 4, "Ashvane Sniper"},
	{138019, 4, "Kul Tiran Vanguard"},
	{137521, 4, "Irontide Powdershot"},
	{138465, 4, "Ashvane Cannoneer"},
	{135241, 4, "Bilge Rat Pillager"},
	{137516, 4, "Ashvane Invader"},
	{141282, 1, "Kul Tiran Footman"},
	{129373, 4, "Dockhound Packmaster"},
	{129371, 4, "Riptide Shredder"},
	{129369, 8, "Irontide Raider"},
	{129370, 4, "Irontide Waveshaper"},
	{138255, 4, "Ashvane Spotter"},
	{129367, 4, "Bilge Rat Tempest"},
	{137517, 4, "Ashvane Destroyer"},
	{138247, 1, "Irontide Marauder"},
	{138254, 1, "Irontide Powdershot"},
	{141565, 1, "Kul Tiran Footman"},
	{141283, 4, "Kul Tiran Halberd"},
	{141565, 1, "Kul Tiran Footman"},
	{129366, 4, "Bilge Rat Buccaneer"},

	--Temple of Sethraliss

	{134600, 4, "Sandswept Marksman"},
	{134616, 2, "Krolusk Pup"},
	{134991, 6, "Sandfury Stonefist"},
	{136076, 6, "Agitated Nimbus"},
	{134629, 6, "Scaled Krolusk Rider"},
	{134599, 4, "Imbued Stormcaller"},
	{134602, 4, "Shrouded Fang"},
	{135562, 2, "Venomous Ophidian"},
	{135846, 2, "Sand-Crusted Striker"},
	{68819, 12, "Eye of Sethraliss"},
	{134990, 4, "Charged Dust Devil"},
	{134617, 1, "Krolusk Hatchling"},
	{139422, 6, "Scaled Krolusk Tamer"},
	{139946, 6, "Heart Guardian"},
	{139949, 4, "Plague Doctor"},
	{134686, 4, "Mature Krolusk"},
	{134364, 4, "Faithless Tender"},
	{139425, 4, "Crazed Incubator"},
	{134691, 4, "Static-charged Dervish"},
	{136250, 4, "Hoodoo Hexer"},

	--The Motherlode
	{133432, 5, "Venture Co. Alchemist"},
	{130436, 1, "Off-Duty Laborer"},
	{136470, 4, "Refreshment Vendor"},
	{130488, 4, "Mech Jockey"},
	{130435, 5, "Addled Thug"},
	{130437, 2, "Mine Rat"},
	{136643, 6, "Azerite Extractor"},
	{134012, 6, "Taskmaster Askari"},
	{133430, 8, "Venture Co. Mastermind"},
	{133963, 1, "Test Subject"},
	{137029, 5, "Ordnance Specialist"},
	{133436, 5, "Venture Co. Skyscorcher"},
	{137940, 4, "Safety Shark"},
	{133345, 5, "Feckless Assistant"},
	{134232, 4, "Hired Assassin"},
	{136139, 6, "Mechanized Peacekeeper"},
	{130661, 8, "Venture Co. Earthshaper"},
	{130653, 4, "Wanton Sapper"},
	{134005, 1, "Shalebiter"},
	{136934, 4, "Weapons Tester"},
	{133593, 5, "Expert Technician"},
	{130485, 6, "Mechanized Peacekeeper"},
	{133482, 1, "Crawler Mine"},
	{137713, 1, "Big Money Crab"},
	{130635, 4, "Stonefury"},
	{133463, 8, "Venture Co. War Machine"},

	--The Underrot
	
	{131402, 1, "Underrot Tick"},
	{131436, 6, "Chosen Blood Matron"},
	{133663, 4, "Fanatical Headhunter"},
	{133852, 4, "Living Rot"},
	{133870, 4, "Diseased Lasher"},
	{131492, 4, "Devout Blood Priest"},
	{133836, 4, "Reanimated Guardian"},
	{134284, 4, "Fallen Deathspeaker"},
	{130909, 4, "Fetid Maggot"},
	{133685, 4, "Befouled Spirit"},
	{133912, 6, "Bloodsworn Defiler"},
	{138187, 4, "Grotesque Horror"},
	{133835, 4, "Feral Bloodswarmer"},
	{138281, 6, "Faceless Corruptor"},

	-- Tol Dagor
	{127477, 6, "Saltwater Snapper"},
	{127480, 1, "Stinging Parasite"},
	{127381, 3, "Silt Crab"},
	{135366, 6, "Blacktooth Arsonist"},
	{135699, 7, "Ashvane Jailer"},
	{127486, 7, "Ashvane Officer"},
	{130027, 7, "Ashvane Marine"},
	{127485, 3, "Bilge Rat Looter"},
	{136665, 7, "Ashvane Spotter"},
	{131112, 6, "Cutwater Striker"},
	{127497, 9, "Ashvane Warden"},
	{127482, 4, "Sewer Vicejaw"},
	{130025, 7, "Irontide Thug"},
	{130026, 6, "Bilge Rat Seaspeaker"},
	{135254, 4, "Irontide Raider"},
	{131112, 6, "Cutwater Striker"},
	{131445, 9, "Block Warden"},
	{127488, 7, "Ashvane Flamecaster"},
	
	--Waycrest Manor
	{131666, 4, "Coven Thornshaper"},
	{131850, 4, "Maddened Survivalist"},
	{135234, 3, "Diseased Mastiff"},
	{131849, 4, "Crazed Marksman"},
	{134024, 1, "Devouring Maggot"},
	{134041, 4, "Infected Peasant"},
	{137830, 4, "Pallid Gorger"},
	{131586, 4, "Banquet Steward"},
	{131685, 4, "Runic Disciple"},
	{131677, 6, "Heartsbane Runeweaver"},
	{135052, 1, "Blight Toad"},
	{131670, 6, "Heartsbane Vinetwister"},
	{131858, 4, "Thornguard"},
	{131819, 4, "Coven Diviner"},
	{131587, 5, "Bewitched Captain"},
	{135329, 8, "Matron Bryndle"},
	{131847, 4, "Waycrest Reveler"},
	{131585, 4, "Enthralled Guard"},
	{135474, 4, "Thistle Acolyte"},
	{135049, 2, "Dreadwing Raven"},
	{131669, 1, "Jagged Hound"},
	{131812, 6, "Heartsbane Soulcharmer"},
	{135365, 8, "Matron Alma"},
	{135240, 2, "Soul Essence"},
	{131818, 4, "Marked Sister"},
	{139269, 4, "Gloom Horror"},
	
	



}

local dbFixes = {}
dbFixes[4] = {{1, 40, "Test NPC"}, {99307, 12, "Skjal"}}
dbFixes[5] = {{2, 35, "Test NPC"}, {95947, 4, "Mak'rana Hardshell"}}

-- New to lua, don't judge too hard pls.

function mppGetLastKill()
	return lastKill
end

--
-- GENERAL ADDON UTILITY
-- And by "utility" I mostly mean creating a bunch of shit that should really be built-in.

local function debugLog(s)
	if mppDebug then
		print(s)
	end
end

local function split(str)
	a = {}
	for s in string.gmatch(str, "%S+") do
		table.insert(a, s)
	end
	return a
end

local function round(number, decimals)
    return (("%%.%df"):format(decimals)):format(number)
end

local function tlen(t)
	local length = 0
	for _ in pairs(t) do
		length = length + 1
	end
	return length
end

local function GetTime()
	return _GetTime() * 1000
end

local function hasWarned(message)
	for _,warning in pairs(warnings) do
		if warning == message then
			return true
		end
	end
	return false
end

local function warning(message)
	if not hasWarned(message) then
		print(message)
		table.insert(warnings, message)
		return true
	end
	return false
end

local function getSetting(setting)
	if (not setting or MythicPlusPullEstimatorDB["settings"][setting] == nil) then
		warning("MPP attempted to get missing setting: " .. setting or "nil")
		return
	end
	return MythicPlusPullEstimatorDB["settings"][setting]
end

function mppGetSetting(setting)
	return getSetting(setting)
end

local function setSetting(setting, value)
	if (not setting or MythicPlusPullEstimatorDB["settings"][setting] == nil) then
		warning("MPP attempted to set missing setting: " .. setting or "nil")
		return
	end
	MythicPlusPullEstimatorDB["settings"][setting] = value
	return value
end

local function toggleSetting(setting)
	return setSetting(setting, not getSetting(setting))
end

--
-- WOW GENERAL WRAPPERS/EZUTILITIES
--

local function getNPCID(guid)
	local targetType, _,_,_,_, npcID = strsplit("-", guid)
	if targetType == "Creature" or targetType == "Vehicle" and npcID then
		return tonumber(npcID)
	end
end

-- TODO: Figure out how to filter out bosses.
local function isValidTarget(targetToken)
	if UnitCanAttack("player", targetToken) and not UnitIsDead(targetToken) then
		return true
	end
end

local function getSteps()
	return select(3, C_Scenario.GetStepInfo())
end

local function isDungeonFinished()
	if mppSimulationMode then return false end
	return (getSteps() and getSteps() < 1)
end

-- Will also return true in challenge modes if those are ever re-implemented as M+ is basically recycled Challenge Mode.
local function isMythicPlus()
	if mppSimulationMode then return true end
	local difficulty = select(3, GetInstanceInfo()) or -1
	if difficulty == 8 and not isDungeonFinished() then
		return true
	else
		return false
	end
end

-- DEPRECATEDa
local function isTeeming()
	warning("Deprecated function isTeeming called!")
	return
end

local function getProgressInfo()
	if isMythicPlus() then
		local numSteps = select(3, C_Scenario.GetStepInfo()) -- GetStepInfo tells us how many steps there are.
		if numSteps and numSteps > 0 then
			local info = {C_Scenario.GetCriteriaInfo(numSteps)} -- It should be the last step.
			if info[13] == true then -- if isWeightedProgress
				return info
			end
		end
	end
end

function mppGetProgressInfo()
	return getProgressInfo()
end
	
local function getMaxQuantity()
	if mppSimulationMode then return simulationMax end
	local progInfo = getProgressInfo()
	if progInfo then
		return getProgressInfo()[5]
	end
end

local function getCurrentQuantity()
	if mppSimulationMode then return simulationCurrent end
	return strtrim(getProgressInfo()[8], "%")
end

local function getEnemyForcesProgress()
	debugLog("getEnemyForcesProgress called.")
	-- Returns exact float value of current enemies killed progress (1-100).
	local quantity, maxQuantity = getCurrentQuantity(), getMaxQuantity()
	local progress = quantity / maxQuantity
	debugLog(progress * 100)
	return progress * 100
end

--
-- DB READ/WRITES
--

local function getValue(npcID)
	debugLog("getValue called. Args: " .. npcID)
	local npcData = MythicPlusPullEstimatorDB["npcData"][npcID]
	if npcData then
		local hiValue, hiOccurrence = nil, -1
		for value, occurrence in pairs(npcData["values"]) do
			if occurrence > hiOccurrence then
				hiValue, hiOccurrence = value, occurrence
			end
		end
		if hiValue ~= nil then
			return hiValue
		end
	end
end

local function deleteEntry(npcID)
	local exists = (MythicPlusPullEstimatorDB["npcData"][npcID] ~= nil)
	MythicPlusPullEstimatorDB["npcData"][npcID] = nil
	return exists
end

local function updateValue(npcID, value, npcName, forceQuantity)
	--debugLog("updateValue called. Args: " .. npcID or "nil" .. ", " .. value or "nil" .. ", " .. npcName or "nil" .. ", " .. forceQuantity or "nil")
	if value <= 0 then
		debugLog("Discarding update for " .. toString(npcName) .. "(" .. npcID .. ") due to value being " .. tovalue)
		return
	end
	local npcData = MythicPlusPullEstimatorDB["npcData"][npcID]
	if not npcData then
		MythicPlusPullEstimatorDB["npcData"][npcID] = {values={}, name=npcName or "Unknown"}
		return updateValue(npcID, value, npcName, forceQuantity)
	end
	local values = npcData["values"]
	if values[value] == nil then
		values[value] = (forceQuantity or 1)
	else
		values[value] = values[value] + (forceQuantity or 1)
	end
	for val, occurrence in pairs(values) do
		if val ~= value then
			values[val] = occurrence * 0.75 -- Newer values will quickly overtake old ones
		end
	end
end

-- Temp testing global access
function mppUpdateValue(npcID, value, npcName, forceQuantity)
	return updateValue(npcID, value, npcName, forceQuantity)
end

-- Temp testing global access
function mppGetValue(npcID, value)
	return getValue(npcID)
end

local function exportData()
	local a = string.format("Export ver %s (%s) - %i mobs: {", getSetting("version"), getSetting("dbVersion"), tlen(MythicPlusPullEstimatorDB["npcData"]))
	for npcID,t in pairs(MythicPlusPullEstimatorDB["npcData"]) do
	   local value = getValue(npcID)
	   local name = t["name"]
	   a = a .. "{".. npcID..","..value..",\""..name.."\"},"
	end
	a = a .. "}"
	local f = CreateFrame('EditBox', "MPPExportBox", UIParent, "InputBoxTemplate")
	f:SetSize(100, 50)
	f:SetPoint("CENTER", 0, 150)
	f:SetScript("OnEnterPressed", f.Hide)
	f:SetText(a)
end

--
-- Light DB wrap
--

-- Returns a nil-100 number representing the percentual progress that npcID is expected to give you.
local function getEstimatedProgress(npcID)
	debugLog("getEstimatedProgress called. Args: " .. npcID)
	local npcValue, maxQuantity = getValue(npcID), getMaxQuantity()
	if npcValue and maxQuantity then
		return (npcValue / maxQuantity) * 100
	end
end

--
-- TRIGGERS/HOOKS
--

-- Called when our enemy forces criteria increases, no matter how small the increase (but >0).
local function onProgressUpdated(deltaProgress)
	debugLog("onProgressUpdated called. Args: " .. deltaProgress)
	if currentQuantity == getMaxQuantity() then
		return -- Disregard data that caps us as we don't know if we got the full value.
	end
	local timestamp, npcID, npcName, isDataUseful = unpack(lastKill) -- See what the last mob we killed was
	if timestamp and npcID and deltaProgress and isDataUseful then -- Assert that we have some useful data to work with
		local timeSinceKill = GetTime() - timestamp
		debugLog("timeSinceKill: " .. timestamp .. " Current Time: " .. GetTime() .. "Timestamp of kill: " .. timeSinceKill)
		if timeSinceKill <= getSetting("maxTimeSinceKill") then
			debugLog(string.format("Gained %f%s. Last mob killed was %s (%i) %fs ago", deltaProgress, "%", npcName, npcID, timeSinceKill/1000))
			updateValue(npcID, deltaProgress, npcName) -- Looks like we have ourselves a valid entry. Set this in our database/list/whatever.
		else
			debugLog(string.format("Gained %f%s. Last mob killed was %s (%i) %fs ago (PAST CUTOFF!)", deltaProgress, "%", npcName, npcID, timeSinceKill))
		end
	end
end

-- Called directly by our hook
local function onCriteriaUpdate()
	debugLog("onCriteriaUpdate called")
	if not currentQuantity then
		currentQuantity = 0
	end
	if not isMythicPlus() or not loaded or not getSetting("enabled") then return end
	newQuantity = getCurrentQuantity()
	deltaQuantity = newQuantity - currentQuantity
	if deltaQuantity > 0 then
		currentQuantity = newQuantity
		onProgressUpdated(deltaQuantity)
	end
		
end

-- Called directly by our hook
local function onCombatLogEvent(args)
	--local _,combatType,_,_,_,_,_, destGUID, destName = unpack(args)
	--if combatType == "UNIT_DIED" then
	local timestamp, combatType, something, srcGUID, srcName, srcFlags, something2, destGUID, destName, destFlags = unpack(args)
	if combatType == "PARTY_KILL" then
		if not isMythicPlus() then return end
		if mppDebug then
			--foreach(args, print)
		end
		local npcID = getNPCID(destGUID)
		if npcID then
			local isDataUseful = true
			local timeSinceLastKill = GetTime() - lastKill[1]
			if timeSinceLastKill <= getSetting("inconclusiveDataThreshold") then
				debugLog("Data not useful: " .. timeSinceLastKill .. " - " .. lastKill[1] .. " - " .. GetTime())
				isDataUseful = false
			end
			lastKill = {GetTime(), npcID, destName, isDataUseful} -- timestamp is not at all accurate, we use GetTime() instead.
			if mppDebug then
				foreach(lastKill, print)
			end
		end
	end
end

local function verifySettings(forceWipe)
	for setting, value in pairs(defaultSettings) do
		if MythicPlusPullEstimatorDB["settings"][setting] == nil or forceWipe then
			MythicPlusPullEstimatorDB["settings"][setting] = value
		end
	end
	-- At last, update version string
	
	setSetting("version", version)
end

local function upgradeDB()
	local oldVer = MythicPlusPullEstimatorDB["settings"]["dbVersion"]
	for ver, fixes in pairs(dbFixes) do
		if ver > oldVer then
			for _,fixTable in pairs(fixes) do
				local npcID, value, name = unpack(fixTable)
				deleteEntry(npcID)
				updateValue(npcID, value, name)
			end
		end
	end
	setSetting("dbVersion", numVersion)
end
				

local function verifyDB(forceWipe)
	if not MythicPlusPullEstimatorDB or not MythicPlusPullEstimatorDB["settings"] or not MythicPlusPullEstimatorDB["npcData"] or forceWipe then
		print("Running first time MPP setup. This should only happen once. Enjoy! ;)")
		MythicPlusPullEstimatorDB = {}
		MythicPlusPullEstimatorDB["settings"] = {}
		MythicPlusPullEstimatorDB["settings"]["dbVersion"] = defaultSettings["dbVersion"]
		MythicPlusPullEstimatorDB["npcData"] = {}
	else
		MythicPlusPullEstimatorDB["settings"]["dbVersion"] = MythicPlusPullEstimatorDB["settings"]["dbVersion"] or 0
	end
	verifySettings()
	if mppDefaultValues ~= nil then
		for k,v in pairs(mppDefaultValues) do
			local npcID, value, name = unpack(v)
			if getValue(npcID) == nil then
				updateValue(npcID, value, name, 1)
			end
		end
	end
	
	-- DB Fixes per version
	upgradeDB()
end

-- Called directly by our hook
local function onAddonLoad(addonName)
	if addonName == "MythicPlusPullEstimator" then
		verifyDB()
		if isMythicPlus() then
			quantity = getEnemyForcesProgress()
			debugLog("MPP Loaded in progress: " .. quantity .. "in.")
		else
			quantity = 0
			debugLog("MPP loaded not in progress.")
		end
		loaded = true
	end
end

---
--- TOOLTIPS
---
	
local function addLineToTooltip(str)
    GameTooltip:AddDoubleLine(str)
    GameTooltip:Show()
end

local function shouldAddTooltip(unit)
	if loaded and getSetting("enabled") and getSetting("enableTooltip") and isMythicPlus() and isValidTarget(unit) then
		return true
	end
	return false
end

local function getTooltipMessage(npcID)
	local tempMessage = "|cFF"..getSetting("tooltipColor").."M+Progress: "
	local estProg = getEstimatedProgress(npcID)
	if not estProg then
		return tempMessage .. "No record."
	end
	mobsLeft = (100 - getEnemyForcesProgress()) / estProg
	tempMessage = string.format("%s%.2f%s (%i left)", tempMessage, estProg, "%", math.ceil(mobsLeft))
	return tempMessage
end
	
local function onNPCTooltip(self)
	local unit = select(2, self:GetUnit())
	if unit then
		local guid = UnitGUID(unit)
		npcID = getNPCID(guid)
		if npcID and shouldAddTooltip(unit) then
			local tooltipMessage = getTooltipMessage(npcID)
			if tooltipMessage then
				addLineToTooltip(tooltipMessage)
			end
		end
	end
end

---
--- SHITTY CURRENT PULL FRAME
---

currentPullFrame = CreateFrame("frame", "currentPullFrame12", UIParent)
mppFrame = currentPullFrame
currentPullFrame:SetPoint("CENTER", UIParent, 400, 300)
currentPullFrame:EnableMouse(true)
currentPullFrame:SetMovable(true)
currentPullFrame:RegisterForDrag("LeftButton")
currentPullFrame:SetScript("OnDragStart", currentPullFrame.StartMoving)
currentPullFrame:SetScript("OnDragStop", currentPullFrame.StopMovingOrSizing)
currentPullFrame:SetWidth(50)
currentPullFrame:SetHeight(50)
--currentPullFrame:SetAllPoints()
--currentPullFrame:SetClampRectInsets(200, 400, 200, 500)
currentPullText = currentPullFrame:CreateFontString("currentPullString", "BACKGROUND", "GameFontHighlightLarge")
currentPullText:SetPoint("CENTER");
currentPullText:SetText("MPP String Uninitialized.")

---
--- NAMEPLATES
---

local function isTargetPulled(target)
	-- debugLog("isTargetPulled with args: " ..target)
	local threat = UnitThreatSituation("player", target) or -1 -- Is nil if we're not on their aggro table, so make it -1 instead.
	if isValidTarget(target) and (threat >= 0 or UnitPlayerControlled(target.."target")) then
		return true
	end
	return false
end
	
local function getPulledUnits()
	local tempList = {}
	for _, nameplate in pairs(C_NamePlate.GetNamePlates()) do
		if nameplate.UnitFrame.unitExists then
			if isTargetPulled(nameplate.UnitFrame.displayedUnit) then
				table.insert(tempList, UnitGUID(nameplate.UnitFrame.displayedUnit))
			end
		end
	end
	return tempList
end

local function getPulledProgress(pulledUnits)
	local estProg = 0
	for _, guid in pairs(pulledUnits) do
		npcID = getNPCID(guid)
		if npcID then
			estProg = estProg + (getEstimatedProgress(npcID) or 0)
		end
	end
	return estProg
end

local function shouldShowCurrentPullEstimate()
	if getSetting("enabled") and getSetting("enablePullEstimate") and isMythicPlus() and not isDungeonFinished() then
		if getSetting("pullEstimateCombatOnly") and not UnitAffectingCombat("player") then
			return false
		end
		return true
	end
	return false
end

local function setCurrentPullEstimateLabel(s)
	currentPullString:SetText(s)
	currentPullFrame:SetWidth(currentPullString:GetStringWidth())
	currentPullFrame:SetHeight(currentPullString:GetStringHeight())
	--print(currentPullFrame:GetCenter())
end

local function updateCurrentPullEstimate()
	if not shouldShowCurrentPullEstimate() then
		currentPullFrame:Hide()
		return
	else
		currentPullFrame:Show()
	end
	local pulledUnits = getPulledUnits()
	local estProg = getPulledProgress(pulledUnits)
	-- debugLog(tlen(pulledUnits) .. "/" .. tlen(activeNameplates).." active nameplates for an estimated " .. estProg .. "%")
	local curProg = getEnemyForcesProgress()
	local totProg = estProg + curProg
	if estProg == 0 then
		tempMessage = "No recorded mobs pulled or nameplates inactive."
	else
		tempMessage = string.format("Current pull: %.2f%s + %.2f%s = %.2f%s", curProg, "%", estProg, "%", (math.floor(totProg*100)/100), "%")
	end
	setCurrentPullEstimateLabel(tempMessage)
end

local function createNameplateText(token)
	local npcID = getNPCID(UnitGUID(token))
	if npcID then
		if activeNameplates[token] then
			activeNameplates[token]:Hide() -- This should never happen...
		end
		local nameplate = C_NamePlate.GetNamePlateForUnit(token)
		if nameplate then
			activeNameplates[token] = nameplate:CreateFontString(token.."mppProgress", nameplate.UnitFrame.healthBar, "GameFontHighlightSmall")
			activeNameplates[token]:SetText("+?%")
		end
	end
end

local function removeNameplateText(token)
	if activeNameplates[token] ~= nil then
		activeNameplates[token]:SetText("")
		activeNameplates[token]:Hide()
		activeNameplates[token] = nil
	end
end
	
local function updateNameplateValue(token)
	local npcID = getNPCID(UnitGUID(token))
	if npcID then
		local estProg = getEstimatedProgress(npcID)
		if estProg and estProg > 0 then
			local tempMessage = "|cFF"..getSetting("nameplateTextColor").."+"
			tempMessage = string.format("%s%.2f%s", tempMessage, estProg, "%")
			activeNameplates[token]:SetText(tempMessage)
			activeNameplates[token]:Show()
			return true
		end
	end
	if activeNameplates[token] then -- If mob dies, a new nameplate is created but not shown, and this ui widget will then not exist.
		activeNameplates[token]:SetText("")
		activeNameplates[token]:Hide()
	end
	return false
end

local function updateNameplateValues()
	for token,_ in pairs(activeNameplates) do
		updateNameplateValue(token)
	end
end

local function updateNameplatePosition(token)
	local nameplate = C_NamePlate.GetNamePlateForUnit(token)
	if nameplate.UnitFrame.unitExists and activeNameplates[token] ~= nil then
		activeNameplates[token]:SetPoint("LEFT", nameplate.UnitFrame.name, "LEFT", nameplate.UnitFrame.name:GetWidth(), 0)
	else
		removeNameplateText(token)
		debugLog("Token " .. token or "nil" .. "does not seem to exist. Why are we trying to update it?")
	end
end

local function shouldShowNameplateTexts()
	if getSetting("enabled") and getSetting("enableNameplateText") and isMythicPlus() and not isDungeonFinished() then
		return true
	end
	return false
end

local function onAddNameplate(token)
	if shouldShowNameplateTexts() then
		createNameplateText(token)
		updateNameplateValue(token)
		updateNameplatePosition(token)
	end
end

local function onRemoveNameplate(token)
	removeNameplateText(token)
	activeNameplates[token] = nil -- This line has been made superflous tbh.
end

local function removeNameplates()
	for token,_ in pairs(activeNameplates) do
		removeNameplateText(token)
	end
end

	
local function updateNameplates()
	if shouldShowNameplateTexts() then
		for token,_ in pairs(activeNameplates) do
			updateNameplatePosition(token)
		end
	else
		removeNameplates()
	end
end

---
--- SET UP HOOKS
---

local frame = CreateFrame("FRAME")
frame:RegisterEvent("ADDON_LOADED")
frame:RegisterEvent("SCENARIO_CRITERIA_UPDATE")
frame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")


--frame:RegisterEvent("NAME_PLATE_CREATED")
frame:RegisterEvent("NAME_PLATE_UNIT_ADDED")
frame:RegisterEvent("NAME_PLATE_UNIT_REMOVED")

function frame:OnEvent(event, ...)
	args={...}
	if event == "ADDON_LOADED" then
		onAddonLoad(args[1])
	elseif event == "SCENARIO_CRITERIA_UPDATE" then
		onCriteriaUpdate()
	elseif event == "COMBAT_LOG_EVENT_UNFILTERED" then
		onCombatLogEvent(args)
	elseif event == "NAME_PLATE_CREATED" then
		--onCreateNameplate(...)
	elseif event == "NAME_PLATE_UNIT_ADDED" then
		onAddNameplate(...)
	elseif event == "NAME_PLATE_UNIT_REMOVED" then
		onRemoveNameplate(...)
	else
		if warning("MythicPlusPullEstimator unhandled event: " .. event) then
			for k,v in ipairs(args) do
				print(k,v)
			end
		end
	end
end

function frame:OnUpdate(elapsed)
	currentPullUpdateTimer = currentPullUpdateTimer + elapsed * 1000 -- Not using milliseconds in 2016? WutFace
	if currentPullUpdateTimer >= getSetting("nameplateUpdateRate") then
		currentPullUpdateTimer = 0
		updateCurrentPullEstimate()
		updateNameplateValues()
	end
	updateNameplates()
end

frame:SetScript("OnEvent", frame.OnEvent)
frame:SetScript("OnUpdate", frame.OnUpdate)
GameTooltip:HookScript("OnTooltipSetUnit", onNPCTooltip)

SLASH_MYTHICPLUSPROGRESS1,SLASH_MYTHICPLUSPROGRESS2,SLASH_MYTHICPLUSPROGRESS3 = "/mpp","/mypp","/mythicplusprogress"

-- TODO: Redo this entire thing that I threw together in 2 min
function SlashCmdList.MYTHICPLUSPROGRESS(args, editbox)
	args = split(args)
	args[1] = string.lower(args[1] or "")
	
	if args[1] == "toggle" then
		print("MythicPlusPullEstimator toggled ".. (toggleSetting("enabled") and "ON" or "OFF"))
		
	elseif args[1] == "reset" then
		currentPullFrame:SetPoint("CENTER", UIParent, 50, 50)
		print("MythicPlusPullEstimator' position reset.")
		
	elseif args[1] == "sim" then
		mppSimulationMode = not mppSimulationMode
		print("MPP simulation mode toggled ".. (mppSimulationMode and "ON" or "OFF"))
		
	elseif args[1] == "debug" then
		mppDebug = not mppDebug
		print("MPP debug toggled ".. (mppDebug and "ON" or "OFF"))
		
	elseif args[1] == "updatevalue" then
		updateValue(getNPCID(UnitGUID("target")), tonumber(args[2]), UnitName("target") .. ("(Manual)"))
		
	elseif args[1] == "getvalue" then
		print(getValue(tonumber(args[2])) or "No Data")
		
	elseif args[1] == "getvalues" then
		local npcData = MythicPlusPullEstimatorDB["npcData"][tonumber(args[2])]
		if npcData then
			print(tostring(npcData["name"]))
			foreach(npcData["values"], print)
		else
			print("No data.")
		end
		
	elseif args[1] == "getsetting" then
		print(getSetting(args[2]))
		
	elseif args[1] == "simmax" then
		simulationMax = tonumber(args[2])
		
	elseif args[1] == "simcur" then
		simulationCurrent = tonumber(args[2])
		
	elseif args[1] == "wipeall" then
		verifyDB(true)
		print("RIP Database.")
		
	elseif args[1] == "wipesettings" then
		verifySettings(true)
		
	elseif args[1] == "tooltip" then
		setSetting("enableTooltip", not getSetting("enableTooltip"))
		print("Toggled tooltips ".. (getSetting("enableTooltip") and "on" or "off"))
		
	elseif args[1] == "currentpull" then
		setSetting("enablePullEstimate", not getSetting("enablePullEstimate"))
		print("Toggled \"current pull\" display ".. (getSetting("enablePullEstimate") and "on" or "off"))
		
	elseif args[1] == "dbinfo" then
		print("Mobs recorded: " .. tlen(MythicPlusPullEstimatorDB["npcData"]))
		
	elseif args[1] == "combatonly" then
		setSetting("pullEstimateCombatOnly", not getSetting("pullEstimateCombatOnly"))
		print("Toggled combat only \"current pull\" display ".. (getSetting("pullEstimateCombatOnly") and "on" or "off"))
		
	elseif args[1] == "version" then
		print("MythicPlusPullEstimator Version: " .. getSetting("version") .. " - " .. "DB Version: " .. getSetting("dbVersion"))
		
	elseif args[1] == "nameplates" then
		print("Nameplate text overlay toggled ".. (toggleSetting("enableNameplateText") and "ON" or "OFF"))
		
	elseif args[1] == "export" then
		exportData()
		print("Data export opened, copy it to clipboard then hit enter to close.")
		
	else
		print("MythicPlusPullEstimator commands:")
		print("/mpp toggle (Toggles ALL addon visibility on/off, it will still record npc data while in Mythic+)")
		print("/mpp reset (reset position of 'current pull' text, it likes to run away)")
		print("/mpp version (displays current version of addon)")
		print("/mpp tooltip (toggles the tooltip functionality on/off)")
		print("/mpp currentpull (toggles the \"current pull\" functionality on/off)")
		print("/mpp dbinfo (Shows how many unique mobs you've recorded)")
		print("/mpp combatonly (Toggles only showing the current pull estimate while you or one of your party members are in combat)")
		print("/mpp nameplates (Toggles nameplate text overlay on/off)")
		print("/mpp wipesettings (Resets all settings to default (but not mob data!))")
	end
end